##   author:martinmohan@yahoo.com date:  21/02/2014
#   scadat is a bash command line interface to the tango scada
#   Copyright (C) <2014>  <Martin Mohan>
#
#   This program is free software; you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program; if not, write to the Free Software Foundation,
#   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA

import os,sys,re,argparse,socket,string
import PyTango,scadat_version

def parse_input():
    parser = argparse.ArgumentParser(description=\
            "Linux command line interface to tango (default: TANGO_HOST=localhost:10000)\n"
            ,formatter_class=argparse.RawTextHelpFormatter)
    #parser.add_argument("host_port",nargs="?",
#        help="e.g. localhost:10000")
    parser.add_argument("-v", action="count",default=0,
            help="increase verbosity using -v,-vv")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-l",action="store_true",
            help="List tango devices\n")
    group.add_argument("-r",nargs='+',
            help="Read attributes from a devices: e.g.\n"
            "scadat -r sys/tg_test/1\n"
            "scadat -r sys/tg_test/1/string_scalar -v")
    group.add_argument("-w",nargs='+',
            help="Write to attributes of a device: e.g.\n"
            "scadat -w sys/tg_test/1\n"
            "scadat -w sys/tg_test/1/string_scalar 'hello world'")
    group.add_argument("-c",nargs='+',
            help="Send command to a device: e.g.\n"
            "scadat -c sys/tg_test/1 -v\n"
            "scadat -c sys/tg_test/1/DevLong 1234\n"
            "scadat -c sys/tg_test/1/State")

    args=parser.parse_args()

    if len(sys.argv)==1:
        print ("scadat -h: For usage\nscadat -v: For info")
        sys.exit(1)
    return args 

def main():
    args=parse_input()
    myscadat=scadat_tango(verbose=args.v)

    if args.l:
        myscadat.ldev()
    elif args.r:
        myscadat.read(args.r)

    elif args.w:
        if len(args.w)==1:
            myscadat.write(args.w[0])
        elif len(args.w)==2:
            myscadat.write(args.w[0],args.w[1])
        else:
            print ("Usage: scadat -w sys/tg_test/1/attr_name <value>")

    elif args.c:
        if len(args.c)==1:
            myscadat.cmd(args.c[0])
        elif len(args.c)==2:
            myscadat.cmd(args.c[0],args.c[1])
        else:
            print ("Usage: scadat -c sys/tg_test/1/cmd <value>")
    elif args.v:
        myscadat.info()
        sys.exit(0) #echo $? will return 0 (pass)

class scadat_tango():
    """set os.environ['TANGO_HOST'] to point to machine required\n\
#        1. host_port=myhost:port (passed by user)\n\
    1. Value returned by environment else localhost:10000\n"""

   # def __init__(self.host=None,verbose=0):
    def __init__(self,verbose=0):
        self.verbose=verbose # global verbose level
 
        #default localhost:10000
        os.environ["TANGO_HOST"]=os.getenv("TANGO_HOST", "localhost:10000")
        # TANGO_HOST set?
        try:
            self.host,self.port=os.environ['TANGO_HOST'].split(":")
            os.environ['TANGO_HOST']=self.host+":"+self.port
#            self.db=PyTango.Database()
        except Exception as e:
            self.handle_exception(e,"Error reading TANGO_HOST "+os.environ['TANGO_HOST'])
            sys.exit(1)

        # port open?
        try:
            self.scan_port(self.host,self.port)
        except AttributeError:
            self.handle_exception(e,"Error scan host: "+self.host+" port: "+self.port)
            sys.exit(1)

        # PyTango Database?
        try:
            self.db=PyTango.Database()
        except AttributeError:
            self.handle_exception(e,"Error reading PyTango.Database() at "+os.environ['TANGO_HOST'])
            sys.exit(1)

    def info(self):
        #"""SCADA specific information on the tango host_port"""
        self.stdout("===============scadat HOST===============")
        self.stdout(socket.gethostname() +" ("+socket.gethostbyname(socket.gethostname())+" )")
        self.stdout("scadat_version "+str(scadat_version.version))
        self.stdout("python.version_info "+str(sys.version))
        self.stdout("PyTango.__version_ "+PyTango.__version__)
        self.stdout("os.uname "+str(os.uname()),2)

        self.stdout("===============scadat TARGET===============")
        self.stdout(os.environ['TANGO_HOST']+" ("+socket.gethostbyname(self.host)+":"+self.port+" )")
#        self.stdout("os:  "+self.host+" ("+socket.gethostbyname(self.host)+" )")

        self.stdout(self.db.get_info(),2)

        host_list=self.db.get_host_list()
        MyHosts="\nHosts:"
        for host in host_list:
            MyHosts=MyHosts+("\t%s"%(host))
        self.stdout(MyHosts,2)
        starter_device=self.db.get_device_exported_for_class("Starter").value_string[0]
        deviceHandle=PyTango.DeviceProxy(starter_device)
        Starter=str(deviceHandle.info())
        self.stdout("Starter:"+starter_device+" "+Starter,2)

        try:
            tango_lib_version=str(deviceHandle.get_tango_lib_version()) # >PyTango 8.1.0
            self.stdout("tango_lib_version "+tango_lib_version,2)
        except Exception as e:
            return True

    def ldev(self):
        """List tango devices\n\
-v - Starter exported devices\n
-vv - All exported devices\n
""https://www.tango-controls.org/static/PyTango/latest/doc/html/contents.html\n\
        """                
        if self.verbose==0:  # Only Starter devices is v==0
            self.get_starter_devices()
            return True

        # if verbose > 0 print all devices
        dev_list=self.db.get_device_exported("*")       # All devices
        dserver="^(dserver*)"
#        print "dev_list %s"%dev_list
        for mydev in dev_list:
#            if not re.match(dserver,mydev): # No dserver/*/*
#                self.stdout(mydev)
            self.stdout(mydev,1)
            myclass=self.db.get_class_for_device(mydev)
            try:
                dev_info=self.db.get_device_info(mydev) #Only works PyTango >= 8.1.0
                self.stdout(mydev+" "+dev_info.ds_full_name,2)
            except Exception as e:
                self.stdout(mydev+" "+myclass,2) #Older PyTango

    def open(self,deviceName):
        """Open deviceName and return deviceHandle or error"""
        try:
            DeviceHandle = PyTango.DeviceProxy(deviceName)
            return DeviceHandle
        except Exception as e:
            self.handle_exception(e,"Cannot Open device "+deviceName+" on "+self.host+":"+self.port)

    def get_device_attr(self,device):
        # ^(.*?)/(.*?)/(.*?)$ # matches nearly anything
        # ^([^<]*)/([^<]*)/([^<]*)$ # matches nearly anything
        #pattern="([a-zA-Z0-9_\.-]+)"
        pattern="([^<]*)"  # match anything - even if it is silly
        dev_pattern="^"+pattern+"/"+pattern+"/"+pattern+"$"
        attr_pattern="^"+pattern+"/"+pattern+"/"+pattern+"/"+pattern+"$"

        try:
            if(bool (re.match(attr_pattern,device))):
                m=re.match(attr_pattern,device)
                deviceName=m.group(1)+"/"+m.group(2)+"/"+m.group(3)
                attrName=m.group(4)
                return (device,deviceName,attrName)
            elif(bool (re.match(dev_pattern,device))):
                return (device,device,None)
#        else:
#            self.stderr("Invalid \'"+device+"\' (use: \'aa/bb/cc\' or \'aa/bb/cc/attrName\')")
#            return (device,None,None)

        except Exception as e:
            self.handle_exception(e,"Invalid name\'"+device)
            return (device,None,None)

    def read(self,devices):
        for device in devices:
            try:
                device,deviceName,attrName=self.get_device_attr(device)
                self.attr_read(deviceName,attrName)
            except Exception as e:
                self.handle_exception(e,"Cannot read device "+device)

    def write(self,device,value=None):
        """Send command to a tango device in format attrName(command) arg value"""
        deviceName="deviceName_unknown"
        attrName="attrName_unknown"
        try:
            device,deviceName,attrName=self.get_device_attr(device)
        except Exception as e:
            self.handle_exception(e,"Cannot write to device "+device)
            self.stdout("Usage: scadat -w "+deviceName+"/<attr_name_w> value",2)
            sys.exit(1)

        if not attrName and not value:
#            self.attr_read(device,attrName=None)
            self.stdout("Usage: scadat -w "+deviceName+"/<attr_name_w> value")
        elif attrName and not value:
            DeviceHandle=self.open(deviceName)
            attr_value=DeviceHandle.read_attribute(attrName)
            self.stdout(attr_value,1)
            self.stdout("Usage: scadat -w "+deviceName+"/"+str(attrName)+" <"+str(attr_value.type)+">")
        else:
            try:         # Try to write attribute:
                DeviceHandle=self.open(deviceName)
                attr_value=DeviceHandle.read_attribute(attrName)
                input_value=self.getValueType(attr_value.type,value)
                attr_value=DeviceHandle.write_attribute(attrName,input_value)
            except Exception as e:
                self.handle_exception(e,"error: DeviceHandle.write_attribute(attrName "+attrName+",input_value"+str(input_value))

            lattrName=deviceName+"/"+str(attrName)
            self.stdout(lattrName+"->"+str(input_value))

    def cmd(self,device,value=None):
        """Send command to a tango device <command> <value>"""
        device,deviceName,cmd_name=self.get_device_attr(device)

        if not cmd_name and not value:
            self.cmd_read_all(deviceName) # also read commands
            self.stdout("Usage: scadat -c "+deviceName+"/cmd_name cmd_value")
        elif cmd_name and not value:
            DeviceHandle=self.open(deviceName)
            cmd_type=DeviceHandle.command_query(cmd_name)
            self.stdout(cmd_type,1)
            if cmd_type.in_type == PyTango.DevVoid:
                self.cmd_write(DeviceHandle,cmd_name,PyTango.DevVoid)
            else:
                self.stdout("Usage: scadat -c "+deviceName+"/"+str(cmd_name)+" <"+str(cmd_type.in_type)+">")
        else:
            try:         # Try to write command:
                DeviceHandle=self.open(deviceName)
                self.cmd_write(DeviceHandle,cmd_name,value)
            except Exception as e:
                self.handle_exception(e,"Error writing cmd "+cmd_name+"->"+value)

    def attr_write(self,DeviceHandle,attrName,value):
        """ write 1 attribute """
        attr_value=DeviceHandle.read_attribute(attrName)
        input_value=self.getValueType(attr_value.type,value)
        try:
            attr_value=DeviceHandle.write_attribute(attrName,input_value)
        except Exception as e:
            self.handle_exception(e,"error: DeviceHandle.write_attribute(attrName "+attrName+",input_value"+str(input_value))
            return
        self.stdout(attrName+"->"+str(input_value))

    def attr_read(self,deviceName,attrName=None): # Displays if attribute writeable
        """ read all attributes """
        DeviceHandle=self.open(deviceName)
#        attr_list=DeviceHandle.get_attribute_list()
        # Get attribute list or one attribute
        if(attrName==None):
            try:
                attr_list=DeviceHandle.get_attribute_list()
            except Exception as e:
                self.handle_exception(e,str(deviceName)+" Cannot get_attr_list")
#                self.handle_exception(e,str(deviceName)+" Cannot get_attr_list",1)
                return()
        else:
            attr_list=[attrName]

#        self.cmd_read_all(deviceName) # also read commands
        for attrName in attr_list:
            lattrName=deviceName+"/"+str(attrName)
#            print "deviceName %s, attrName %s"%(deviceName,attrName)
            try:
                attr_value=DeviceHandle.read_attribute(attrName)

                if self.verbose==0:
                    if attr_value.w_value==None:
                        self.stdout(lattrName+" <attr_ro>")
                    else:
                        self.stdout(lattrName+" <attr_w>")
                elif self.verbose==1:
                    self.stdout(lattrName+" "+str(attr_value.value),1)

                self.stdout(lattrName+" "+str(attr_value),2)
            except Exception as e:
                self.handle_exception(e,"Exception read_attribute "+lattrName,1)

    def cmd_write(self,DeviceHandle,cmd_name,value):
        """cmd_write and convert value to correct float, int ,string etc.."""
        cmd_type=DeviceHandle.command_query(cmd_name)
        input_value=self.getValueType(cmd_type.in_type,value)
        self.stdout("self.cmd_write cmd_type.in_type "+str(cmd_type.in_type)+" value "+str(value)+" ->input_value "+str(input_value),2)
        cmd_value=DeviceHandle.command_inout(cmd_name,input_value)
        self.stdout(cmd_value)
        self.stdout("DeviceHandle.command_inout: cmd_name "+cmd_name+", input_value "+str(input_value)+" -> cmd_value "+str(cmd_value)+" -> cmd_type.in_type "+str(cmd_type.in_type),2)

    def cmd_read_all(self,deviceName):
        """ read all cmds """
        DeviceHandle=self.open(deviceName)

        try:
            cmd_list=DeviceHandle.command_list_query() # all
            for cmd in cmd_list:
#                self.stdout(deviceName+"/"+cmd.cmd_name+" "+str(cmd.in_type)+" <cmd>")
                self.stdout(deviceName+"/"+cmd.cmd_name+" <"+str(cmd.in_type)+">")
                self.stdout(cmd,1)
        except Exception as e:
            self.handle_exception(e,str(DeviceHandle)+" :Cannot get cmd_list")

    def get_starter_devices(self):
        """'Read device Starter e.g. tango/admin/scadat)"""
        starter_device="unknown_starter"
        try:
            starter_device=self.db.get_device_exported_for_class("Starter").value_string[0]
#            print "starter_device %s"%starter_device
            deviceHandle = PyTango.DeviceProxy(starter_device)
            RunningServers=deviceHandle.read_attribute("RunningServers").value

            devicelist=[]
            for servers in RunningServers:
                serverName=servers.split("/")[0] # Simulator/test -> Simulator
                self.stdout(serverName,2)
                deviceNames=self.db.get_device_exported_for_class(serverName)
                for deviceName in deviceNames:
                    # Avoid double deviceNames
                    if not any(deviceName in s for s in devicelist):
                        self.stdout(deviceName)
                        devicelist.append(deviceName)
        except Exception as e:
            self.handle_exception(e,"Error reading "+starter_device+"/RunningServers from "+self.host+":"+self.port)
#            self.verbose=1 # If Starter fails show all devices
#            self.ldev()

    def scan_port(self,host,port):
        try:
            host=socket.gethostbyname(host)
            hostname=socket.getfqdn(host)
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(2)
        except Exception as e:
            self.handle_exception(e,"scan_port:Error reading "+os.environ['TANGO_HOST'])
            return

        myconnect = sock.connect_ex((host,string.atoi(port)))
        ipaddr=hostname+":"+port+" ("+(host+":"+port)+")"
#        if myconnect == 0:
#            self.stdout(ipaddr+" open")
#        else:
#            self.stdout(ipaddr+" not open")
        if myconnect != 0:
            self.stdout(ipaddr+" not open")

    # need to transform values for writing?
    def getValueType(self,cmd_type,value):
#        self.out("getValueType cmd_type %s value %s"%(cmd_type,value))
        if cmd_type==PyTango.DevFloat:
            return float(value)
        elif cmd_type==PyTango.DevDouble:
            return float(value)
        elif cmd_type==PyTango.DevLong:
            return long(value)
        elif cmd_type==PyTango.DevLong64:
            return long(value)
        elif cmd_type==PyTango.DevULong64:
            return long(value)
        elif cmd_type==PyTango.DevULong:
            return long(value)
        elif cmd_type==PyTango.DevInt:
            return int(value)
        elif cmd_type==PyTango.DevBoolean:
            if value==str(0):
                return False
            else:
                return True
        elif cmd_type==PyTango.DevShort:
            return long(value)
        elif cmd_type==PyTango.DevUShort:
            return long(value)
        elif cmd_type==PyTango.DevString:
            return str(value)
        elif cmd_type==PyTango.DevVarLongArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarLong64Array:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarULong64Array:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarFloatArray:
            return map(float, value.split())
        elif cmd_type==PyTango.DevVarDoubleArray:
            return map(float, value.split())
        elif cmd_type==PyTango.DevVarShortArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarUShortArray:
            return map(long, value.split())
        elif cmd_type==PyTango.DevVarStringArray:
            return map(str, value.split())
        elif cmd_type==PyTango.DevVoid:
            return None
# Not sure stuff below
        elif cmd_type==PyTango.DevVarDoubleStringArray:
            return "Not tested yet"
#            return map(str, value.split())
        elif cmd_type==PyTango.DevVarCharArray:
#           "1 2 3 4" -> [1 2 3 4]
            return "Not tested yet"
        elif cmd_type==PyTango.DevVarLongStringArray:
            return "Not tested yet"
        elif cmd_type==PyTango.DevState:
            return "Not tested yet"
        else:
            return "Unknown cmd_type %s"%cmd_type

# print if v >= that supplied by user
    def stdout(self,string_value,v=0):
#        if self.verbose==v:
#            print (string_value)
        if v<=self.verbose:
            print (string_value)

    def stderr(self,string_value,v=0):
#        if self.verbose==v:
#            sys.stderr.write(string_value+"\n")
        if v<=self.verbose:
            sys.stderr.write(string_value+"\n")

    ## Use v for more verbose error handling
    def handle_exception(self,e,message="Exception",v=0):
        """ Handle exception"""
#        if self.verbose>=1:
#            print "In verbose1"+str(e)
#            self.stderr(message+"\n"+str(e),v)
#        else:
#            self.stderr(message,v)
        if v<=self.verbose:
            self.stderr(message,v)
